"""Identity abstraction."""
import pprint

from cryptography.hazmat.primitives import serialization

from .dane import DANE
from .pki import PKI
from .exceptions import TLSAError


class Identity:
    """Represent a DANE identity."""

    def __init__(self, dnsname, private_key=None, resolver_override=None, dns_timeout=5):
        """Initialize with the DNS name.

        Args:
            dnsname (str): DNS name of identity.
            private_key (str): Private key in PEM format. Optional.
            resolver_override (str): Override the default resolver IP address.
            dns_timeout (int): Set DNS timeout.

        Raise:
            TLSAError if identity does not exist in DNS.
        """
        self.dnsname = dnsname
        self.dnssec = False
        self.tls = False
        self.tcp = False
        self.private_key = PKI.load_private_key(private_key)
        self.resolver_override = resolver_override
        self.dane_credentials = []
        self.dns_timeout = dns_timeout
        self.set_dane_credentials(self.dnsname)

    def validate_certificate(self, certificate):
        """Return True, None if the certificate is valid for the identity.
        
        This method returns two values, success and status.

        This method only checks against TLSA records with
        certificate_usage 4, or PKIX-CD.
        
        Args:
            certificate (str): Certificate in PEM or DER format.
        
        Returns:
            bool: True if successful, False if validation fails.
            str: Status indicating why validation passed or failed.
        """
        cert_obj = PKI.build_x509_object(certificate)
        why_not = []
        default = "Unable to find a TLSA record with certificate usage 4."
        # For each TLSA certificate, attempt to validate local cert.
        for credential in self.dane_credentials:
            valid = False
            cert_usage = credential["certificate_usage"]
            if cert_usage == "PKIX-CD":
                valid, reason = self.validate_pkix_cd(cert_obj, credential)
                if valid:
                    return True, reason
                else:
                    why_not.append(reason)
        if not why_not:
            why_not.append(default)
        return False, "\n".join(why_not)

    def validate_pkix_cd(self, cert_obj, credential):
        """Validate a certificate with certificate_usage 4.
        
        PKIX-CD expects selector 0 and matching type 0. This
        method will not validate configuration which differs 
        from this expectation.

        Args:
            cert_obj (cryptography.x509): Certificate object.
            credential (dict): Parsed credential from DNS.

        Returns:
            bool: True or False for validation
            string: Reason for validation pass/fail.
        """
        why_not = []
        # Check TLSA records for wrong selector and matching type.
        selector = credential["tlsa_parsed"]["selector"]
        matching_type = credential["tlsa_parsed"]["matching_type"]
        cert_association = credential["tlsa_parsed"]["certificate_association"]
        if selector != 0:
            why_not.append("Selector set to {}.".format(selector))
        if matching_type != 0:
            why_not.append("Matching type set to {}.".format(matching_type))
        if why_not:
            return False, "\n".join(why_not)
        # Check to see that the DER matches what's in DNS
        cert_der = cert_obj.public_bytes(encoding=serialization.Encoding.DER)
        tlsa_der = PKI.certificate_association_to_der(cert_association)
        if not cert_der == tlsa_der:
            return False, "Certificate and TLSA certificate association do not match."
        # Get the CA certificate
        try:
            ca_pems = DANE.get_ca_certificates_for_identity(self.dnsname, cert_der, 100, self.resolver_override, self.dns_timeout)
        except ValueError as err:
            return False, str(err)
        cert_pem = cert_obj.public_bytes(serialization.Encoding.PEM)
        validated, reason = PKI.validate_certificate_chain(cert_pem, ca_pems)
        if not validated:
            return False, "Validation against CA certificate failed: {}.".format(reason)
        return True, "Format and authority CA signature verified."

    def get_pkix_cd_trust_chain(self, certificate, max_levels=100):
        """Return a dictionary with entire discovered trust chain.
        
        Args:
            certificate (str): EE certificate to begin trust chain discovery with.
            max_levels (int): Maximum number of parent certificates to discover. Default: 3.
        
        Returns:
            dict: Dictionary with integer keys for entity cert (``0``) and intermediate CA certificates.
                The root certificate key is ``root``.
        """
        certificate = PKI.build_x509_object(certificate).public_bytes(serialization.Encoding.PEM)
        retval = {0: certificate}
        next_level = 1
        ca_certificates = DANE.get_ca_certificates_for_identity(self.dnsname, certificate, max_levels, self.resolver_override, self.dns_timeout)
        chain_valid, reason = PKI.validate_certificate_chain(certificate, ca_certificates)
        if not chain_valid:
            raise ValueError(reason)
        for cert in ca_certificates:
            aki = PKI.get_authority_key_id_from_certificate(cert)
            ski = PKI.get_subject_key_id_from_certificate(cert)
            if aki == ski:
                # The root cert is the last, so we break here.
                retval["root"] = cert
                break
            retval[next_level] = cert
            next_level += 1
        return retval

    def get_first_entity_certificate(self, strict=True):
        """Return the first entity certificate for the identity.

        Keyword args:
            strict (bool): Raise TLSAError if certificate was not retrieved
                with the benefit of DNSSEC, or in the case of PKIX-CD, if the
                certificate can not be validated via PKI.
        
        Raise:
            TLSAError: If strict is set to ``True`` and the certificate cannot
                be validated by carrying a DNSSEC RRSIG. If ``certificate_usage`` 
                is set to ``4``, PKIX validation may be attempted in lieu of 
                DNSSEC.
            ValueError: If ``cert_type`` is unsupported.
        
        Return:
            cryptography.x509.Certificate: Certificate object as parsed 
                from TLSA record.
        """
        target = ""
        # Find a matching credential
        for cred in self.dane_credentials:
            if not cred["tlsa_parsed"]["matching_type"] == 0:
                continue
            target = cred if cred["tlsa_parsed"]["certificate_usage"] in [1, 3, 4] else ""
            if target:
                break
        if not target:
            raise TLSAError("No entity certificate found for {}.".format(self.dnsname))
        if strict:
            try:
                target["tlsa_parsed"]["dnssec"] = self.dnssec
                DANE.authenticate_tlsa(self.dnsname, target["tlsa_parsed"], self.resolver_override, self.dns_timeout)
            except ValueError as err:
                raise TLSAError(err)
        return target["certificate_object"]


    def get_first_entity_certificate_by_type(self, cert_type, strict=True):
        """Return the first certificate of ``cert_type`` for the identity.
        
        Supported certificate types:
            PKIX-EE: Corresponds with ``certificate_usage`` ``1``.
            DANE-EE: Corresponds with ``certificate_usage`` ``3``.
            PKIX-CD: Corresponds with ``certificate_usage`` ``4``.


        Keyword args:
            strict (bool): Raise TLSAError if certificate was not retrieved
                with the benefit of DNSSEC, or in the case of PKIX-CD, if the
                certificate can not be validated via PKI.
        
        Raise:
            TLSAError: If strict is set to ``True`` and the certificate cannot
                be validated by carrying a DNSSEC RRSIG. If ``certificate_usage`` 
                is set to ``4``, PKIX validation may be attempted in lieu of 
                DNSSEC.
            ValueError: If ``cert_type`` is unsupported.
        
        Return:
            cryptography.x509.Certificate: Certificate object as parsed 
                from TLSA record.
        """
        supported_certificate_types = {"PKIX-EE": 1, "DANE-EE": 3, "PKIX-CD": 4}
        target = ""
        # Verify that we're asked for something legitimate
        if cert_type not in supported_certificate_types:
            raise ValueError("Unsupported cert type {}".format(cert_type))
        type_id = supported_certificate_types[cert_type]
        # Find a matching credential
        for cred in self.dane_credentials:
            if not cred["tlsa_parsed"]["matching_type"] == 0:
                continue
            target = cred if cred["tlsa_parsed"]["certificate_usage"] == type_id else ""
            if target:
                break
        if not target:
            return ""
        if strict:
            try:
                target["tlsa_parsed"]["dnssec"] = self.dnssec
                DANE.authenticate_tlsa(self.dnsname, target["tlsa_parsed"], self.resolver_override, self.dns_timeout)
            except ValueError as err:
                raise TLSAError(err)
        return target["certificate_object"]

    def report(self):
        """Return a report for the identity.

        Prints the query context (DNSSEC, etc) as well as information about
        the TLSA records stored at the identity's name.
        """
        fmt = "Name: {}\n".format(self.dnsname)
        fmt += ("Request context:\n DNSSEC: {}\n TLS: {}\n "
                "TCP: {}\n".format(self.dnssec, self.tls, self.tcp))
        cred_index = 0
        fmt += ("Public credentials: {}\n".format(len(self.dane_credentials)))
        for cert in self.dane_credentials:
            validation_err = ""
            cert["tlsa_parsed"]["dnssec"] = self.dnssec
            try:
                DANE.authenticate_tlsa(self.dnsname, cert["tlsa_parsed"])
            except ValueError as err:
                validation_err = "\n        ".join(["    {}".format(x) for x in str(err).splitlines()])
            validation_status = ("\n    Cryptographically validated." 
                                 if not validation_err 
                                 else "\n  Not validated:\n{}".format(validation_err))
            fmt += ("Credential index: {}\n"
                    "validation status: {} \n"
                    " certificate usage: {}\n"
                    " selector: {}\n"
                    " matching type: {}\n".format(cred_index,
                                                  validation_status,
                                                  cert["certificate_usage"],
                                                  cert["selector"],
                                                  cert["matching_type"]))
            if "certificate_metadata" in cert:
                attributes = pprint.pformat(cert["certificate_metadata"])
                fmt += (" x509 attributes:\n")
                for attribute in attributes.splitlines():
                    fmt += ("    {}\n".format(attribute))
            fmt += "----------"
            fmt += "\n"
        return fmt

    def get_all_certificates(self, filters=[]):
        """Return a dictionary of all EE certificates for this identity.

        This method uses available methods for validating certificates retrieved
        from TLSA records associated with the identity's DNS name. 
        
        For DANE-EE, we really just care that it parses and it was delivered with 
        DNSSEC. 
        
        For PKIX-EE, we require delivery to be protected by DNSSEC. In the future,
        when the Python cryptography library supports full PKIX validation, we will
        also include PKIX validation. https://github.com/pyca/cryptography/issues/2381

        For PKIX-CD, we require that the trust chain be represented out-of-band in 
        accordance with the proposed standard for certificate and trust chain discovery.

        Keyword args:
            filters (list): List of filters for specific DANE certificate usages.
                Valid filters are: "DANE-EE", "PKIX-EE", "PKIX-CD".

        Return: 
            dict: Dictionary key is ``${DNSNAME}-${CERTHASH}``, and the value is the
                the PEM-encoded certificate.
        """
        retval = {}
        # Bail if a bad filter is used.
        if filters:
            for filter_val in filters:
                if filter_val not in ["DANE-EE", "PKIX-EE", "PKIX-CD"]:
                    raise ValueError("Invalid filter: {}".format(filter_val))
        else:
            filters = ["DANE-EE", "PKIX-EE", "PKIX-CD"]
        # Iterate and authenticate
        for cred in self.dane_credentials:
            tlsa = cred["tlsa_parsed"]
            # If it's not a full cert, skip
            if not tlsa["matching_type"] == 0:
                continue
            id_name = self.dnsname 
            cert_obj = cred["certificate_object"]
            cert_pem = cert_obj.public_bytes(serialization.Encoding.PEM)
            cert_hash = DANE.generate_sha_by_selector(cert_pem, "sha256", 0)
            # Validate for PKIX-CD
            if (tlsa["certificate_usage"] == 4 and "PKIX-CD" in filters):
                valid, _ = self.validate_pkix_cd(cert_obj, cred)
                if not valid:
                    continue
                retval["{}-{}".format(id_name, cert_hash)] = cert_pem
            # Validate for DANE-EE (delivered via DNSSEC?)
            if (tlsa["certificate_usage"] == 3 and "DANE-EE" in filters):
                if not self.dnssec:
                    continue
                retval["{}-{}".format(id_name, cert_hash)] = cert_pem
            # Validate for PKIX-EE (delivered via DNSSEC?)
            if (tlsa["certificate_usage"] == 1 and "PKIX-EE" in filters):
                if not self.dnssec:
                    continue
                retval["{}-{}".format(id_name, cert_hash)] = cert_pem
        return retval


    def set_dane_credentials(self, dnsname):
        """Get public credentials from DNS and set DNS retrieval context.

        Args:
            dnsname (str): Name of DNS-based identity.
            resolver_override (str): Optional. Override the default resolver
                IP address.
        """
        tlsa_records = DANE.get_tlsa_records(dnsname, self.resolver_override, self.dns_timeout)
        request_context_fields = ["dnssec", "tcp", "tls"]
        for field in request_context_fields:
            setattr(self, field, tlsa_records[0][field])
        self.dane_credentials = [DANE.process_tlsa(record) for record
                                 in tlsa_records]

    def cert_matches_private_key(self, cert_obj):
        """Return boolean for alignment between private key and cert_obj, and a reason.
        
        Args:
            cert_obj (cryptography.x509): A certificate object.

        Returns:
            bool: True if the public key in the certificate matches the private key.
            str: If validation failed, why it failed.
        """
        cert_public_key = cert_obj.public_key()
        public_key_from_privkey = self.private_key.public_key()
        cert_pubkey_type = type(cert_public_key)
        privkey_pubkey_type = type(public_key_from_privkey)
        if not cert_pubkey_type == privkey_pubkey_type:
            reason = "Key type mismatch: cert: {} privkey: {}.".format(cert_pubkey_type,
                                                                       privkey_pubkey_type)
            return False, reason
        if not (cert_public_key.public_bytes(serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo) 
                == public_key_from_privkey.public_bytes(serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo)):
            reason = "Public key bytes mismatch."
            return False, reason
        return True, ""

    


"""Download all PKIX-CD CA certificates for an identity."""
import argparse
import os

from dane_discovery.dane import DANE
from dane_discovery.pki import PKI
from dane_discovery.identity import Identity



parser = argparse.ArgumentParser(description="Retrieve and store all CA certificates required for PKIX-CD authentication of an identity.")
parser.add_argument("--output_path", dest="out_path", required=True, help="Output path for certificate bundle")
parser.add_argument("--separate_files", dest="separate_files", required=False, 
                    action="store_true", help=("This will use --output_path as"
                        "a directory for writing individual certificate files. "
                        "Individual CA certificate files will be named AUTHORITY_HOSTNAME-CA-subjectKeyID.crt.pem"))
parser.add_argument("--identity_name", dest="dnsname", required=True, help="Identity DNS name")
parser.set_defaults(separate_files=False)

def main():
    """Wrap functionality provided by Identity.get_all_certificates()"""
    # Parse args
    args = parser.parse_args()
    # Get PKIX-CD certs from DNS
    identity = Identity(args.dnsname)
    ee_certs = identity.get_all_certificates(filters=["PKIX-CD"])
    # Get the CA certificates for the EE certs
    certs = {}
    for _, ee_cert_pem in ee_certs.items():
        try:
            identity = Identity(args.dnsname)
            ca_pem = identity.get_pkix_cd_trust_chain(ee_cert_pem)["root"]
        except ValueError as err:
            print(err)
            continue
        except KeyError as err:
            print("Key Error! Are we able to obtain the root CA certificate? {}".format(err))
            continue
        authority_hostname = DANE.generate_authority_hostname(args.dnsname)
        ca_cert_skid = PKI.get_subject_key_id_from_certificate(ca_pem)
        ca_cert_name = "{}-CA-{}".format(authority_hostname, ca_cert_skid.replace("-", ""))
        certs[ca_cert_name] = ca_pem
    # Write out files
    if args.separate_files:
        write_individual_certs(certs, args.out_path)
    else:
        write_cert_bundle(certs, args.out_path)


def write_individual_certs(certs, out_path):
    prepped = [(os.path.join(out_path, "{}.crt.pem".format(name)), contents) 
               for name, contents in certs.items()]
    for item in prepped:
        write_file(*item)


def write_cert_bundle(certs, out_path):
    just_certs = [y for _, y in certs.items()]
    contents = b"\n".join(just_certs)
    write_file(out_path, contents)


def write_file(file_path, contents):
    """Write a string to a file."""
    with open(file_path, "wb") as f_obj:
        f_obj.write(contents)


if __name__ == "__main__":
    main()


"""Exceptions defined here."""


class TLSAError(Exception):
    """General TLSA error, details in ``self.message``."""

class PKIXError(Exception):
    """General PKIX error, details in ``self.message``."""



"""PKI- and Certificate-oriented utilities here."""
import binascii

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.exceptions import InvalidSignature
from cryptography.exceptions import UnsupportedAlgorithm

from .exceptions import TLSAError

class PKI:
    @classmethod
    def build_x509_object(cls, certificate):
        """Return a cryptography.x509.Certificate object.

        Args:
            certificate (str): Certificate in PEM or DER format.

        Return:
            cryptography.x509.Certificate object.

        Raise:
            TLSAError if unable to parse.
        """
        if isinstance(certificate, str):
            certificate = certificate.encode()
        try:
            x5cert = cls.clean_up_certificate(certificate)
            return x509.load_pem_x509_certificate(x5cert,
                                                  default_backend())
        except ValueError:  # This hits if it's a DER cert.
            pass
        return x509.load_der_x509_certificate(certificate,
                                              default_backend())

    @classmethod
    def build_public_key_object_from_der(cls, der):
        """Return a Python cryptography public key object."""
        return serialization.load_der_public_key(der)

    @classmethod
    def serialize_cert(cls, certificate, fmt):
        """Return certificate bytes in the selected format.
        
        Args:
            certificate (cryptography.x509.Certificate): Certificate to parse.
            fmt (str): DER, PEM, or RPK_DER. RPK_DER is raw public key, DER encoding.
        
        Returns:
            bytes: Serialized certificate.
            
        Raises:
            ValueError if an invalid format was requested.
        """
        if fmt == "PEM":
            return certificate.public_bytes(serialization.Encoding.PEM)
        elif fmt == "DER":
            return certificate.public_bytes(serialization.Encoding.DER)
        elif fmt == "RPK_DER":
            pubkey = certificate.public_key()
            return pubkey.public_bytes(serialization.Encoding.DER,
                                       serialization.PublicFormat.SubjectPublicKeyInfo)
        raise ValueError("Unsupported serialization format requested.")

    @classmethod
    def certificate_association_to_der(cls, certificate_association):
        """Return DER bytes from a TLSA record's ``certificate_association``.

        Args:
            certificate_association (str): Certificate association information
                extracted from a TLSA record.

        Return:
            bytes: DER-formatted certificate.
        """
        return binascii.unhexlify(certificate_association)
    
    @classmethod
    def clean_up_certificate(cls, certificate):
        """This method returns a clean PEM-encoded certificate.
        
        This is useful for removing the human-readable certificate
        metadata that sometimes ends up in certificates produced 
        by OpenSSL.
        """
        if isinstance(certificate, bytes):
            certificate = certificate.decode("utf-8")
        in_lines = certificate.splitlines()
        out_lines = []
        for line in in_lines:
            if line.startswith(" "):
                continue
            if line.startswith("Certificate:"):
                continue
            out_lines.append(line)
        return "\n".join(out_lines).encode()

    @classmethod
    def der_to_pem(cls, der_cert):
        """Return the PEM representation of a TLSA certificate_association.

        Args:
            der (str): A certificate in DER format.

        Return:
            bytes: PEM-formatted certificate.
        """
        cert = cls.build_x509_object(der_cert)
        return cert.public_bytes(serialization.Encoding.PEM)

    @classmethod
    def format_keyid(cls, keyid):
        """Return dash-delimited string from keyid."""
        delimiter = '-'
        hex_bytes = []
        keyid = [x for x in keyid]
        keyid.reverse()
        while keyid:
            first = keyid.pop()
            second = keyid.pop()
            hex_bytes.append("{}{}".format(first, second))
        return delimiter.join([x for x in hex_bytes])
    
    @classmethod
    def get_authority_key_id_from_certificate(cls, certificate):
        """Extract and return the authorityKeyIdentifier from the certificate.
        
        Args:
            certificate (str): Certificate in PEM or DER format.
        """
        cert_obj = cls.build_x509_object(certificate)
        akid = cert_obj.extensions.get_extension_for_class(
            x509.AuthorityKeyIdentifier
            ).value.key_identifier
        akid_hex = binascii.hexlify(akid).decode()
        return cls.format_keyid(akid_hex)

    @classmethod
    def get_cert_meta(cls, cert_der):
        """Return a dictionary containing certificate metadata."""
        retval = {"subject": {}, "extensions": {}}
        x509_obj = cls.build_x509_object(cert_der)
        for item in x509_obj.subject:
            retval["subject"][item.oid._name] = item.value
        for extension in x509_obj.extensions:
            xtn = cls.parse_extension(extension)
            xtn_name = [x for x in xtn.keys()][0]
            retval["extensions"][xtn_name] = xtn[xtn_name]
        return retval

    @classmethod
    def get_dnsnames_from_cert(cls, x5_obj):
        """Return the dnsnames from the certificate's SAN.

        Args:
            x5_obj (cryptography.x509): Certificate object.
        
        Return: 
            list: str: dNSNames from certificate SAN.
        """
        san = x5_obj.extensions.get_extension_for_class(x509.SubjectAlternativeName)
        return san.value.get_values_for_type(x509.DNSName)

    @classmethod
    def get_subject_key_id_from_certificate(cls, certificate):
        """Return the subjectKeyIdentifier for the certificate.

        Args:
            certificate (str): Certificate in PEM or DER format.
        """
        cert_obj = cls.build_x509_object(certificate)
        skid = x509.SubjectKeyIdentifier.from_public_key(cert_obj.public_key())
        skid_hex = binascii.hexlify(skid.digest).decode()
        return cls.format_keyid(skid_hex)

    @classmethod
    def is_a_ca_certificate(cls, certificate):
        """Return True if ``certificate`` is a CA certificate."""
        x5_obj = cls.build_x509_object(certificate)
        basic_constraints = x5_obj.extensions.get_extension_for_class(x509.BasicConstraints)
        if basic_constraints.value.ca is True:
            return True
        return False

    @classmethod
    def load_private_key(cls, private_key):
        """Return a private key.

        Wraps cryptography.hazmat.primitives.serialization.load_pem_private_key
        and returns the appropriate type.

        Args:
            private_key (str): Private key in PEM format.

        Return:
            Private key object, or None if ``private_key`` arg is None.

        Raise:
            ValueError if there's an error loading the key.
        """
        if private_key is None:
            return None
        return serialization.load_pem_private_key(private_key, password=None)

    @classmethod
    def parse_extension(cls, x509_ext):
        """Return a dictionary representation of the x509 extension."""
        usage = ["content_commitment", "crl_sign", "data_encipherment",
                 "digital_signature", "key_agreement", "key_cert_sign",
                 "key_encipherment"]
        oidname = x509_ext.oid._name
        if oidname == "basicConstraints":
            return {"BasicConstraints": {"ca": x509_ext.value.ca,
                                        "path_length":
                                        x509_ext.value.path_length}}
        elif oidname == "extendedKeyUsage":
            return {"ExtendedKeyUsage": [oid._name for oid
                                         in x509_ext.value]}
        elif oidname == "subjectAltName":
            return {"SubjectAltName": [oid.value for oid in x509_ext.value]}
        elif oidname == "keyUsage":
            return {"KeyUsage": {x: getattr(x509_ext.value, x) for x in usage}}
        return {x509_ext.oid._name: x509_ext.value}

    @classmethod
    def validate_certificate_association(cls, certificate):
        """Raise TLSAError if certificate association is not a certificate or public key, or return None.

        Args:
            certificate (str): Certificate association data from TLSA record.

        Return:
            None

        Raise:
            TLSAError if parsing fails.
        """
        try:
            # First, we try to parse as x509
            der = cls.certificate_association_to_der(certificate)
            cls.build_x509_object(der)
            return
        except (ValueError, binascii.Error) as err:
            pass
        try:
            # Next, we try as raw public key
            der = cls.certificate_association_to_der(certificate)
            cls.build_public_key_object_from_der(der)
            return
        except (ValueError, binascii.Error) as err:
            pass
        raise TLSAError("The certificate association does not parse as a certificate or raw public key.")

    @classmethod
    def validate_certificate_chain(cls, entity_certificate, ca_certificates):
        """Return True if PKI trust chain is established from entity to CA.
        
        This method attempts cryptographic validation of ``entity_certificate`` 
        against the list of ``ca_certificates``. This method only checks 
        public keys and signatures, independent of any x509v3 extensions.

        The validation process completes successfully if a self-signed CA
        certificate is encountered in ``ca_certificates``, which terminates a
        cryptographically-validated chain from the entity certificate.

        Args:
            entity_certificate (str): Entity certificate to be verified.
            ca_certificates (list of str): List of CA certificates for validating
                ``entity_certificate``.
        
        returns:
            (True, None) if certificate validates.
            (False, str) if certificate does not validate, and str will contain the reason.
        """
        validation = {cls.get_subject_key_id_from_certificate(c): c
                      for c in ca_certificates}
        ca_skis = [x for x in validation.keys()]
        for v in ca_skis:
            if not PKI.is_a_ca_certificate(validation[v]):
                print("Rejecting non-ca certificate with SKI {}".format(v))
                del(validation[v])
        # Set the initial cert for chain validation to the entity certificate
        currently_validating = x509.load_pem_x509_certificate(entity_certificate, default_backend()).public_bytes(serialization.Encoding.PEM)
        if not validation:
            return (False, "No CA certificates supplied!")
        # While we still have certs in the list...
        while validation:
            current_aki = cls.get_authority_key_id_from_certificate(currently_validating)
            # If we don't have a CA cert that matches the current entity's AKI, we bail.
            if current_aki not in validation:
                msg = "Parent certificate not found!"
                return (False, msg)
            identified_parent = validation[current_aki]
            # If the signature doesn't match, bail.
            if not cls.verify_certificate_signature(currently_validating, identified_parent):
                msg = "Cryptographic certificate validation failed!"
                return (False, msg)
            # If the parent CA certificate is self-signed...
            if cls.get_authority_key_id_from_certificate(identified_parent) == cls.get_subject_key_id_from_certificate(identified_parent):
                return (True, None)
            currently_validating = x509.load_pem_x509_certificate(validation[current_aki], default_backend()).public_bytes(serialization.Encoding.PEM)
            del(validation[current_aki])
        msg = "Unable to build trust chain to self-signed CA certificate."
        return (False, msg)

    @classmethod
    def verify_certificate_signature(cls, certificate, ca_certificate):
        """Return True if certificate was signed by ca_certificate.

        Args:
            entity_certificate (str): entity certificate in DER or PEM format.
            ca_certificate (str): CA certificate in DER or PEM format.

        Return: 
            bool: True if the ca_certificate validates the entity_certificate.
        """
        issuer_public_key = cls.build_x509_object(ca_certificate).public_key()
        cert_to_check = cls.build_x509_object(certificate)
        if isinstance(issuer_public_key, RSAPublicKey):
            try:
                issuer_public_key.verify(cert_to_check.signature,
                    cert_to_check.tbs_certificate_bytes, padding.PKCS1v15(),
                    cert_to_check.signature_hash_algorithm)
            except InvalidSignature:
                return False
        elif isinstance(issuer_public_key, EllipticCurvePublicKey):
            try:
                issuer_public_key.verify(cert_to_check.signature,
                    cert_to_check.tbs_certificate_bytes,
                    cert_to_check.signature_hash_algorithm)
            except InvalidSignature:
                return False
            except UnsupportedAlgorithm:
                return False
        else:
            raise ValueError("Unsupported public key type {}".format(type(issuer_public_key)))
        return True

    @classmethod
    def verify_dnsname(cls, dns_name, certificate):
        """Return True if the first dNSName in the SAN matches."""
        x5_obj = cls.build_x509_object(certificate)
        san_dns_names = cls.get_dnsnames_from_cert(x5_obj)
        if san_dns_names[0] != dns_name:
            return False
        return True


"""Download all PKIX-CD certificates for an identity."""
import argparse
import os

from dane_discovery.identity import Identity



parser = argparse.ArgumentParser(description=("Retrieve, authenticate, and store all certificates for a DANE identity.\n"
                                 "Default behavior retrieves and authenticates all available entity certificates."
                                 "Adding filters for specific types (--pkix-cd, for instance) limits output to those types."))
parser.add_argument("--output_path", dest="out_path", required=True, help="Output path for certificate bundle")
parser.add_argument("--separate_files", dest="separate_files", required=False, 
                    action="store_true", help=("This will use --output_path as"
                        "a directory for writing individual certificate files."))
parser.add_argument("--identity_name", dest="dnsname", required=True, help="Identity DNS name")
# Filters by type
parser.add_argument("--dane_ee", dest="filter_dane_ee", required=False, action="store_true", help="Include DANE-EE.")
parser.add_argument("--pkix_ee", dest="filter_pkix_ee", required=False, action="store_true", help="Include PKIX-EE.")
parser.add_argument("--pkix_cd", dest="filter_pkix_cd", required=False, action="store_true", help="Include PKIX-CD.")
parser.set_defaults(separate_files=False, filter_dane_ee=False, filter_pkix_ee=False, filter_pkix_cd=False)

def main():
    """Wrap functionality provided by Identity.get_all_certificates()"""
    # Parse args
    args = parser.parse_args()
    filters = []
    if args.filter_dane_ee:
        filters.append("DANE-EE")
    if args.filter_pkix_ee:
        filters.append("PKIX-EE")
    if args.filter_pkix_cd:
        filters.append("PKIX-CD")
    # Get PKIX-CD certs from DNS
    identity = Identity(args.dnsname)
    certs = identity.get_all_certificates(filters=filters)
    # Write out files
    if args.separate_files:
        write_individual_certs(certs, args.out_path)
    else:
        write_cert_bundle(certs, args.out_path)


def write_individual_certs(certs, out_path):
    prepped = [(os.path.join(out_path, "{}.crt.pem".format(name)), contents) 
               for name, contents in certs.items()]
    for item in prepped:
        write_file(*item)


def write_cert_bundle(certs, out_path):
    just_certs = [y for _, y in certs.items()]
    contents = b"\n".join(just_certs)
    write_file(out_path, contents)


def write_file(file_path, contents):
    """Write a string to a file."""
    with open(file_path, "wb") as f_obj:
        f_obj.write(contents)


if __name__ == "__main__":
    main()


"""Authenticate a certificate using DANE PKIX-CD mode."""
import argparse
import sys

from dane_discovery.pki import PKI
from dane_discovery.identity import Identity


parser = argparse.ArgumentParser(description="Authenticate a local certificate using PKIX-CD")
parser.add_argument("--certificate_path", dest="cert_path", required=True, help="Path to certificate")
parser.add_argument("--identity_name", dest="dnsname", required=False, help="Identity DNS name")
parser.add_argument("--silent", dest="silent", action="store_true", help="No output, only exit code")
parser.set_defaults(silent=False, dnsname=None)


def main():
    """Wrap functionality provided by Identity.validate_certificate()"""
    # Parse args
    args = parser.parse_args()
    # Load cert from file
    try:
        cert_file = get_file_contents(args.cert_path)
    except Exception as err:
        exit_handler(False, str(err), args.silent)
    # Determine DNS name
    try:
        if not args.dnsname:
            cert_obj = PKI.build_x509_object(cert_file)
            dns_name = PKI.get_dnsnames_from_cert(cert_obj)[0]
        else:
            dns_name = args.dnsname 
    except IndexError:
        exit_handler(False, "No DNS name via args or certificate.", args.silent)
    # Run validator
    identity = Identity(dns_name)
    success, msg = identity.validate_certificate(cert_file)
    # Return results
    exit_handler(success, msg, args.silent)


def get_file_contents(file_path):
    """Get the contents of a file."""
    with open(file_path) as f_obj:
        return f_obj.read()


def exit_handler(success, msg, silent):
    """Handle exiting gracefully."""
    if not silent:
        print(msg)
    if success:
        sys.exit(0)
    exit(1)


if __name__ == "__main__":
    main()